/* SDL handling
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include "print.h"
#include "sdl.h"
#include "opengl.h"

#define GL3_PROTOTYPES 1
#include <GL/glew.h>

static int sdl_initialized = 0;
static int audio_initialized = 0;
static SDL_Window *gl_window = NULL;
static SDL_GLContext gl_context = NULL;
static void (*keyboard_sdl)(int down, SDL_Keycode sym) = NULL;
static void (*audio_sdl)(float *data, int len) = NULL;

static void audio_cb(void __attribute__((unused)) *userdata, Uint8 *stream, int len)
{
	float audio_data[len / sizeof(float)];

	SDL_memset(stream, 0, len);
	audio_sdl(audio_data, len / sizeof(float));
	SDL_MixAudio(stream, (Uint8 *)audio_data, len, SDL_MIX_MAXVOLUME);
}

int init_sdl(const char *progname, int width, int height, int sound_samplerate, int sound_chunk, void (*keyboard)(int down, SDL_Keycode sym), void (*audio)(float *data, int len))
{
	int rc;

	keyboard_sdl = keyboard;
	audio_sdl = audio;

	/* init SDL library */
	rc = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);
	if (rc < 0) {
		print_error("Failed to init SDL\n");
		goto error;
	}
	sdl_initialized = 1;

	/* open window */
	gl_window = SDL_CreateWindow(progname, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, width, height, SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE);
	if (!gl_window) {
		print_error("Failed to open SDL window: %s\n", SDL_GetError());
		rc = EIO;
		goto error;
	}

	/* create GL context */
	gl_context = SDL_GL_CreateContext(gl_window);
	if (!gl_context) {
		print_error("Failed to create SDL's OpenGL context: %s\n", SDL_GetError());
		rc = EIO;
		goto error;
	}

	rc = SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
	if (rc < 0) {
		print_error("Failed to set SDL's OpenGL context profile\n");
		goto error;
	}

	rc = SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
	if (rc < 0) {
		print_error("Failed to set SDL's OpenGL major version\n");
		goto error;
	}
	rc = SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
	if (rc < 0) {
		print_error("Failed to set SDL's OpenGL minor version\n");
		goto error;
	}

	rc = SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	if (rc < 0) {
		print_error("Failed to set SDL's OpenGL doublebuffer\n");
		goto error;
	}

	rc = SDL_GL_SetSwapInterval(1);
	if (rc < 0) {
		print_error("Failed to set SDL's OpenGL swap interval to VBLANK\n");
		goto error;
	}

#ifndef __APPLE__
	glewExperimental = GL_TRUE;
	if (glewInit() != GLEW_OK) {
		print_error("Failed to init GLEW\n");
		goto error;
	}
#endif

	/* just in case */
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	/* open audio */
	SDL_AudioSpec want, have;
	SDL_memset(&want, 0, sizeof(want)); /* or SDL_zero(want) */
	want.freq = sound_samplerate;
	want.format = AUDIO_F32; /* we always use float in this project */
	want.channels = 1;
	want.samples = sound_chunk; /* must be a power of two */
	want.callback = audio_cb;
	rc = SDL_OpenAudio(&want, &have);
	if (rc < 0) {
		print_error("Failed to open audio\n");
		goto error;
	} else if (have.format != want.format) {
		print_error("Failed to open audio with desired audio format\n");
		SDL_CloseAudio();
		rc = -EIO;
		goto error;
	} else if (have.freq != want.freq) {
		print_error("Failed to open audio with desired sample rate\n");
		SDL_CloseAudio();
		rc = -EIO;
		goto error;
	} else {
		SDL_PauseAudio(0);
		audio_initialized = 1;
	}

	return 0;

error:
	exit_sdl();
	return rc;
}

static int key_ctrl = 0, fullscreen = 0;

int event_sdl(void)
{
	int quit = 0;
	SDL_Event event;

	while (SDL_PollEvent(&event)) {
		if (event.type == SDL_QUIT)
			quit = 1;
		if (event.type == SDL_WINDOWEVENT) {
			if (event.window.event == SDL_WINDOWEVENT_SIZE_CHANGED)
				resize_opengl(event.window.data1, event.window.data2);
		}
		if (event.type == SDL_KEYDOWN) {
			switch (event.key.keysym.sym) {
			case SDLK_f:
				if (key_ctrl) {
					if (fullscreen) {
						fullscreen = 0;
						SDL_SetWindowFullscreen(gl_window, 0);
					} else {
						SDL_SetWindowFullscreen(gl_window, SDL_WINDOW_FULLSCREEN_DESKTOP);
						fullscreen = 1;
					}
				}
				break;
			case SDLK_LCTRL:
			case SDLK_RCTRL:
				key_ctrl = 1;
				break;
			}
			keyboard_sdl(1, event.key.keysym.sym);
		}
		if (event.type == SDL_KEYUP) {
			switch (event.key.keysym.sym) {
			case SDLK_LCTRL:
			case SDLK_RCTRL:
				key_ctrl = 0;
				break;
			}
			keyboard_sdl(0, event.key.keysym.sym);
		}
	}

	return quit;
}

void swap_sdl(void)
{
	SDL_GL_SwapWindow(gl_window);
}

void exit_sdl(void)
{
	/* close window */
	if (gl_window) {
		SDL_DestroyWindow(gl_window);
		gl_window = NULL;
	}

	/* clear OpenGL context */
	if (gl_context) {
		SDL_GL_DeleteContext(gl_context);
		gl_context = NULL;
	}

	/* exit SDL library */
	if (audio_initialized) {
		SDL_PauseAudio(1);
		SDL_CloseAudio();
		audio_initialized = 0;
	}
	if (sdl_initialized) {
		SDL_Quit();
		sdl_initialized = 0;
	}
}

